#!/usr/bin/perl
#
# This script uses the Tk perl module to generate a plot of
# the current traffic pattern seen by a specified sensor.
#
# Two kinds of plots are possible.  The first (default) plot is a circle
# of host names with lines connecting them indicating communication.
# Green lines are ICMP, red are UDP and blue are TCP.  The width of the
# line connecting two hosts is proportional to the volume of traffic
# between them.  An arrowhead indicates the general direction of the
# flow of data.  I.e., 1.2.3.4 -> 5.6.7.8 means that more than half of
# the data between the two hosts was sent by 1.2.3.4 to 5.6.7.8 .
# If you've ever seen etherman, this plot should look fairly familiar.
#
# The other sort of plot (obtained by giving the -p flag on the command
# line) creates a logarithmic plot of number of bytes of traffic per
# second versus time in seconds.
#
# Placing the mouse cursor over addresses or plot lines should highlight
# the selected object. The status window above the plot should update to 
# contain the address(es) of the selected object.
#
# Clicking on a highlighted object will bring up a window showing the
# traffic associated with that object (either the traffic to and from
# a single host, or the traffic between two hosts).
#
# Widget buttons:
#
#	Pause		Pauses the plot.  The script will continue to collect
#			data, and any host windows will continue to update.
#	Close		Closes the input pipe.  The plot will remain, but no
#			additional data will be received
#	Quit		Exits the script
#
# Command line flags:
#
#	-p		Display a logarithmic plot rather than a circle
#			of hosts
#	-r [dump]	Read data from dumpfile [dump]
#	-s [ip]		Connect (via ssh(1)) to sensor with address [ip]
#	-l [luser]	Lusername for ssh session
#	-i [identity]	Ssh(1) identity file to use
#

#use strict;

use POSIX;
use POSIX":sys_wait_h";
use Tk;
use Tk::Balloon;
use Tk qw/:eventtypes/;
use vars qw/$main %backfills $opt_A $opt_i $opt_l $opt_w $opt_s $opt_x $opt_y $opt_R %old_hosts %old_list $radius $range $window $max_x $max_y %hostlist %pipelist %hosts %proto $text $bottom $luser $identity $host $paused %pipes %protos $pi *PIPE $epoch $opt_r $ssh $alert %colours %hcolours @hcolour $closed $lastsec $plot_font $scrollbuffer $status $foo $balloon/;

require("getopts.pl");

&Getopts("A:i:l:pw:s:x:y:w:r:R:");

$| = 1;

$left_margin = 50;
$right_margin = 50;
$top_margin = 50;
$bottom_margin = 50;

$plot_font = '-*-Helvetica-Medium-R-Normal--*-SIZE-*-*-*-*-*-*';
#@highlight = (qw/-foreground=>'#888888'/);
$ssh = '/usr/local/bin/ssh';
$alert = '/usr/local/net/ids/bin/alert';
$scrollbuffer = 10;
%colours =
(
	'icmp',		'green',
	'tcp',		'blue',
	'udp',		'red',
);
%protos =
(
	1,	'icmp',
	6,	'tcp',
	17,	'udp',
);
%hcolours = (
	0,	'purple',
	1,	'red',
	2,	'green',
	3,	'blue',
	4,	'brown',
	5,	'orange',
	6,	'yellowgreen',
	7,	'skyblue1',
);
@hcolour = keys %hcolours;
$pi = atan2(1,1) * 4;

&main();
exit(0);


sub main
{
	my($time0, $time1, $foo);
	&do_config();
	$main = new MainWindow(-title=>"Traffic Monitor Widget");
	&open_pipe();
	&open_window();
	if($opt_r)
	{
		$time0 = time;
		while(1)
		{
			$time1 = time;
			DoOneEvent(DONT_WAIT);
			$paused && next;
			($time1 > ($time0 + $pauselen)) &&
				(($time0 = time), $epoch = 0);
			$closed && next;
			$epoch && next;
			$foo = <PIPE>;
			&read_data($foo);
		}
	}
	else
	{
		$main->fileevent(PIPE, 'readable', [\&read_data]);
	}
	MainLoop;
}

sub open_pipe
{
	$lastsec = 0;
	($opt_r) && (open(PIPE, "$alert -r $opt_r -K|") ||
		(die "Couldn't open input pipe"), return);
	open(PIPE, "$ssh -i $identity -l $luser $host|") ||
		(die "Couldn't open input pipe");
}

sub print_sec
{
	my($sec) = @_;
	my($foo, $src, $dst, $i, $max_size);
	($paused != 0) && return;
	$text->delete(qw/1.0 end/);
	foreach $foo (sort(keys %pipes))
	{
		$pipes{$foo} || next;
		($pipes{$foo} > $max_size) && ($max_size = $pipes{$foo});
		$i = int($pipes{$foo});
		($src, $dst) = split('\s+', $foo);
		$i && 
			$text->insert('end', "$src\t$dst\t\t$i\n");
	}
	$text->yview('end');
	$opt_p && (&do_plot($sec), 1) || &do_circle();
	$opt_r && $epoch++;
	$lastsec = $sec;
	%pipes = ();
	%hosts = ();
	%proto = ();
}

sub open_window
{
	my($close);
	my($text_yscroll, $text_xscroll, $top, $frame1);
	$top = $main->Frame;
	$frame1 = $main->Frame;
	$frame2 = $main->Frame;
	$bottom = $main->Canvas(-width=>$max_x, -height=>$max_y)->pack(-side=>'bottom', -fill=>'both');
	$top->Label(-text=>'Traffic Monitor Widget')->pack(-side=>'left');
	$top->pack;
	$text = $frame1->Text(-width=>80, -height=>10, -wrap=>'none');
	$text_xscroll = $frame1->Scrollbar(-orient=>'horiz',
		-command=>['xview', $text])->pack;
	$text->configure(-xscrollcommand=>['set', $text_xscroll]);
	$text_yscroll = $frame1->Scrollbar(-orient=>'vert',
		-command=>['yview', $text])->pack;
	$text->configure(-yscrollcommand=>['set', $text_yscroll]);
	$text_xscroll->pack(-side=>'bottom', -fill=>'x');
	$text_yscroll->pack(-side=>'right', -fill=>'y');
	$text->pack(-expand=>1, -side=>'left');
	$top->Button(-text=>'Pause',
		-command=>sub{\&do_pause()})->pack(-side=>'right');
	$close = $top->Button(-text=>'Close',
		-command=>sub{$closed++;close(PIPE);$close->destroy})->pack(-side=>'right');
	$top->Button(-text=>'Quit',
		-command=>sub{close(PIPE);exit(0)})->pack(-side=>'right');
	$status = $frame2->Text(-width=>60, -height=>1, -relief=>"sunken",
		-bd=>1);
	$status->pack(-side=>"bottom", -fill=>"y", -padx=>2, -pady=>1);
	#$balloon = $frame2->Balloon(-statusbar => $status);
	$frame1->pack;
	$frame2->pack;
	$bottom->pack;
}

sub read_data
{
	my($baz) = @_;
	my(@foo, $bar, $blah, $tmp, $src, $dst, $sport, $dport, $size);
	($opt_r && eof) && ($closed++, close(PIPE));
	$opt_r && (($bar = $baz), 1) || ($bar = <PIPE>);
	foreach $baz (keys %hostlist)
	{
		($bar =~ /$baz/) || next;
		$blah = "text_$baz";
		$tmp = $bar;
		$tmp =~ s/\r//g;
		${$blah}->insert('end', "$tmp");
		${$blah}->yview('end');
	}
	foreach $baz (keys %pipelist)
	{
		($src, $dst) = split('#', $baz);
		(($bar =~ /$src/) && ($bar =~ /$dst/)) || next;
		$blah = "text_$baz";
		$tmp = $bar;
		$tmp =~ s/\r//g;
		${$blah}->insert('end', "$tmp");
		${$blah}->yview('end');
	}
	#print "$bar";
	@foo = split('\s+', $bar);
	($lastsec == 0) && ($lastsec = $foo[1]);
	($lastsec < ($foo[1] - $window))  && &print_sec($foo[1]);
	$src = $foo[2];
	($src =~ s/:(\d+)//g) && ($sport = $1);
	$src =~ s/:://g;
	$dst = $foo[4];
	($dst =~ s/:(\d+)//g) && ($dport = $1);
	$dst =~ s/:://g;
	$size = $foo[6];
	$pipes{"$src $dst"} += $size;
	$pipes{"$dst $src"} += 0.000001;
	$pipes{"total"} += $size;
	$hosts{"$src"} += $size;
	$hosts{"$dst"} += 0.000001;
	$hosts{"total"} += $size;
	$proto{"$src $dst"} = $protos{$foo[5]};
	$proto{"$dst $src"} = $protos{$foo[5]};
	$kept_hosts{"$src"}++;
	$kept_hosts{"$dst"}++;
	$kept_hosts{"total"}++;
}

sub do_plot
{
	my($now) = @_;
	my($max_host, $sec, $port, $host, $old_list, %ports);
	my($first, $last, $i, @list, $colour, $xscale, $yscale);
	my(%allhosts, $x, $y, $src, $dst, $font, $sq, @point, $foo);

	$bottom->delete('all');
	$bottom->createRectangle(0, 0, $max_x, $max_y, -fill=>'black', -width=>0, -outline=>undef);
	$sec = int($now - ($window / 2));
	$old_list{$sec} = "";
	foreach $host (sort(keys %kept_hosts))
	{
		$old_hosts{$sec}{$host} = $hosts{$host};
		$old_list{$sec} .= "$host#";
	}
	($lastsec == 0) && return;
	$old_list{$sec} =~ s/#$//g;
	$first = int($sec - $range);
	$last = int($sec);
	$i = $first;
	$max_host = 0;
	while($i <= $last)
	{
		#($old_list{$i} eq "") && ($i++, next);
		@list = split('#', $old_list{$i});
		foreach $host (@list)
		{
			$old_hosts{$i}{$host} && $allhosts{$host}++;
			($max_host < $old_hosts{$i}{$host}) && ($max_host = $old_hosts{$i}{$host});
		}
		$i++;
	}
	$xscale = ($max_x - $left_margin - $right_margin) /
		(($last - $first) / $window);
	($xscale < 1) && ($xscale = 1);
	$yscale = ($max_y - $top_margin - $bottom_margin) /
		(int(&my_log10($max_host) + 1));
	for($i = $first;  $i <= $last; $i += $window)
	{
		$x = int($xscale * (($i - $first) / $window)) + $left_margin;
		$bottom->createLine($x, $bottom_margin, $x, ($max_y - $top_margin),, -width=>1, -fill=>'grey');
	}
	$font = $plot_font;
	$font =~ s/SIZE/90/g;
	for($i = $bottom_margin; $i <= ($max_y - $top_margin); $i += $yscale)
	{
		$bottom->createLine($left_margin, $i, ($max_x - $right_margin), $i,, -width=>1, -fill=>'grey');
		$j = 10 ** (int((($max_y - $i - $top_margin - $bottom_margin)
			/ $yscale) + 0.5) + 1);
		$bottom->createText(2, $i, -text=>"$j", -font=>$font, -fill=>'white', -anchor=>'w');
	}
	@list = keys(%allhosts);
	foreach $host (@list)
	{
		($src, $dst) = split('#', $host);
		@point = ();
		for($i = $first;  $i <= $last; $i += $window)
		{
			$x = int($xscale * (($i - $first) / $window));
			$y = (&my_log10($old_hosts{$i}{$host}) * $yscale);
			($y < 0) && ($y = 0);
			push(@point, int($x + $left_margin), ($max_y - $y - $top_margin));
		}
		#$bottom->createLine(@point, -width=>1, -fill=>'white');
		$colour = $host;
		$colour =~ s/\D+//g;
		$bottom->createLine(@point, -width=>1, -fill=>$hcolours{$colour % $#hcolour}, tag=>$host);
		$bottom->bind($host, '<1>'=>[\&click_pipe]);
		$bottom->bind($host, '<Enter>'=>sub{\&highlight_host("$src#$dst")});
		$bottom->bind($host, '<Leave>'=>sub{\&unhighlight_host("$src#$dst")});
	}
}

sub do_circle
{
	my($foo, $space, $a, $i, $rradius);
	my($x, $y, $rx, $ry, $ox, $oy, $label, @labels, $count);
	my($entries, $colour, $max_host, $backfill);
	my($xmiddle, $ymiddle, $xz, $yz, $anchor, $tag, %done, $arrow);
	my(%lx, %ly, $src, $dst, $sq, $font, $max_size);

	%lx = ();
	%ly = ();
	$bottom->delete('all');
	$bottom->createRectangle(0, 0, $max_x, $max_y, -fill=>'black', -width=>0, -outline=>undef);
	$xmiddle = ($max_x / 2);
	$ymiddle = ($max_y / 2);
	$xz = ($max_x / 16);
	$yz = ($max_y / 16);
	$entries = 0;
	@labels = ();
	foreach $foo (sort(keys %hosts))
	{
		($foo eq 'total') && next;
		($hosts{$foo} > $max_host) && ($max_host = $hosts{$foo});
		push(@labels, $foo);
		$entries++;
	}
	$rradius = $radius;
	($entries < 2) && return;
	$space = &deg2rad(360 / $entries);
	while(($space * ($rradius / $radius)) < .1)
	{
		$rradius++;
	}
	for($i = 1;  $i <= $entries; $i++)
	{
		$x = $rradius * cos($a);
		$y = $rradius * sin($a);
		$rx = $xmiddle + $x;
		$ry = $ymiddle + $y;
		$font = $plot_font;
		$font =~ s/SIZE/120/g;
		$label = shift(@labels);
		$label || next;
		$anchor = '';
		($ry < ($ymiddle - $yz)) && ($anchor = 's');
		($ry > ($ymiddle + $yz)) && ($anchor = 'n');
		($rx < ($xmiddle - $xz)) && ($anchor .= 'e');
		($rx > ($xmiddle + $xz)) && ($anchor .= 'w');
		$anchor || ($anchor = 'center');
		$label && $bottom->createText($rx, $ry, -text=>"$label", -font=>$font, -fill=>'white', -anchor=>$anchor, -tags=>'label');
		$bottom->bind('label', '<1>'=>[\&click_host]);
		$backfill = "back_$label";
		$bottom->bind('label', '<Enter>'=>sub{\&highlight_host($label)});
		$bottom->bind('label', '<Leave>'=>sub{\&unhighlight_host($label)});
		$lx{$label} = $rx;
		$ly{$label} = $ry;
		$a += $space;
	}
	%done = ();
	foreach $foo (sort(keys %pipes))
	{
		($pipes{$foo} > $max_size) && ($max_size = $pipes{$foo});
	}
	foreach $foo (sort(keys %pipes))
	{
		($src, $dst) = split('\s+', $foo);
		($done{"$src $dst"} != 0) && next;
		$done{"$src $dst"}++;
		$done{"$dst $src"}++;
		if($pipes{"$src $dst"} < $pipes{"$dst $src"})
		{
			$count = $pipes{"$dst $src"};
			$arrow = 'first';
		}
		else
		{
			$count = $pipes{"$src $dst"};
			$arrow = 'last';
		}
		$colour = $colours{$proto{"$src $dst"}};
		$colour || ($colour = 'white');
		$sq = int((10 / $max_size) * $count) + 1;
		($lx{$src} && $ly{$src} && $lx{$dst} && $ly{$dst}) || next;
		$bottom->createLine($lx{$src}, $ly{$src}, $lx{$dst}, $ly{$dst}, -width=>$sq, -fill=>$colour, -tags=>"$src#$dst", -arrow=>$arrow, -capstyle=>'round');
		$bottom->bind("$src#$dst", '<1>'=>[\&click_pipe]);
		$bottom->bind("$src#$dst", '<Enter>'=>sub{\&highlight_host("$src#$dst")});
		$bottom->bind("$src#$dst", '<Leave>'=>sub{\&unhighlight_host("$src#$dst")});
	}
	%done = ();
	$opt_r && $epoch++;
}

sub deg2rad
{
	my($deg) = @_;
	return($deg*$pi/180);
}

sub do_pause
{
	$paused && (($paused = 0), 1) || ($paused++);
	$epoch = 0;
	$paused && ($text->insert('end', "(Paused)\n"), 1) ||
		($text->insert('end', "(Unpaused)\n"));
	$text->yview('end');
	return($paused);
}

sub click_host
{
	my($id, $foo, $bar, $baz, $blah, $tmp, $sx, $sy);
	$foo = ($bottom->itemconfigure('current', -text))[4];
	$hostlist{$foo}++;
	$bar = "win_$foo";
	${$bar} && return;
	${$bar} = $main->Toplevel(-title=>"$foo");
	$tmp = "button_$foo";
	${$tmp} = ${$bar}->Frame;
	${$tmp}->pack;
	${$tmp}->Button(-text=>'Close',
		-command=>sub{delete $hostlist{$foo};${$bar}->destroy})->pack(-side=>'right');
	$baz = "frame_$foo";
	${$baz} = ${$bar}->Frame;
	${$baz}->pack;
	$blah = "text_$foo";
	${$blah} = ${$baz}->Text(-width=>80, -height=>25);
	$sx = "xs_$foo";
	$sy = "ys_$foo";
	${$sx} = ${$baz}->Scrollbar(-orient=>'horiz',
		-command=>['xview', ${$blah}]);
	${$sy} = ${$baz}->Scrollbar(-orient=>'vert',
		-command=>['yview', ${$blah}]);
	${$blah}->configure(-xscrollcommand=>['set', ${$sx}]);
	${$blah}->configure(-yscrollcommand=>['set', ${$sy}]);
	${$sx}->pack(-side=>'bottom', -fill=>'x');
	${$sy}->pack(-side=>'right', -fill=>'y');
	${$blah}->pack;
}

sub click_pipe
{
	my($foo, $bar, $baz, $blah, $tmp, $sx, $sy, @t, $label);
	$bar = ($bottom->itemconfigure('current', -tags))[4];
	$foo = $bar->[0];
	$label = $foo;
	$label =~ s/#/ > /g;
	$pipelist{$foo}++;
	$bar = "win_$foo";
	${$bar} && return;
	${$bar} = $main->Toplevel(-title=>"$label");
	$tmp = "button_$foo";
	${$tmp} = ${$bar}->Frame;
	${$tmp}->pack;
	${$tmp}->Button(-text=>'Close',
		-command=>sub{delete $pipelist{$foo};${$bar}->destroy})->pack(-side=>'right');
	$baz = "frame_$foo";
	${$baz} = ${$bar}->Frame;
	${$baz}->pack;
	$blah = "text_$foo";
	${$blah} = ${$baz}->Text(-width=>80, -height=>25);
	$sx = "xs_$foo";
	$sy = "ys_$foo";
	${$sx} = ${$baz}->Scrollbar(-orient=>'horiz',
		-command=>['xview', ${$blah}]);
	${$sy} = ${$baz}->Scrollbar(-orient=>'vert',
		-command=>['yview', ${$blah}]);
	${$blah}->configure(-xscrollcommand=>['set', ${$sx}]);
	${$blah}->configure(-yscrollcommand=>['set', ${$sy}]);
	${$sx}->pack(-side=>'bottom', -fill=>'x');
	${$sy}->pack(-side=>'right', -fill=>'y');
	${$blah}->pack;
}

sub highlight_host
{
	my($backfill) = @_;
	my($foo, $bar);
	$backfills{$backfill} = ($bottom->itemconfigure('current', -fill))[4];
	$bottom->itemconfigure('current', -fill=>'yellow');
	$host =~ s/#/ /g;
	if($backfill =~ /#/)
	{
		($bar = ($bottom->itemconfigure('current', -tags))[4]);
		$foo = $bar->[0];
		$foo =~ s/#/ > /g;
	}
	else
	{
		($foo = ($bottom->itemconfigure('current', -text))[4]);
	}
	$status->delete(qw/1.0 end/);
	$status->insert('end', $foo);
}

sub unhighlight_host
{
	my($backfill) = @_;
	$bottom->itemconfigure('current', -fill=>$backfills{$backfill});
}

sub do_config
{
	$identity = $opt_i;
	$luser = $opt_l;
	$host = $opt_s;
	$window = $opt_w;
	$max_x = $opt_x;
	$max_y = $opt_y;
	$radius = $opt_R;
	$range = $opt_A;
	$range || ($range = 450);
	$radius || ($radius = 150);
	$window || ($window = 10);
	$pauselen = ($window / 10);
	$max_x || ($max_x = 750);
	$max_y || ($max_y = 750);
	#$opt_r && (($epoch = 1), ($paused = 1), return);
	$opt_r && (($epoch = 1), return);
	$host || (die "You need to specify a sensor with -s");
	$luser || (die "You need to specify a lusername with -l");
	$identity || (die "You need to specify an identity file with -i");
}

sub my_log10
{
	my($foo) = @_;
	($foo == 0) && ($foo = 1);
	return(log($foo)/log(10));
}
